/* wordsplit - a word splitter
   Copyright (C) 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program. If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>

#include <smap/wordsplit.h>

#ifdef HAVE_GETTEXT_H
# include <gettext.h>
#else
# define gettext(s) s
#endif
#define _(msgid) gettext(msgid)
#define N_(msgid) msgid

#define ISWS(c) ((c)==' '||(c)=='\t'||(c)=='\n')
#define ISDELIM(c,delim) (strchr(delim,(c))!=NULL)
#define ISPUNCT(c) (strchr("!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~",(c))!=NULL)
#define ISUPPER(c) ('A' <= ((unsigned) (c)) && ((unsigned) (c)) <= 'Z')
#define ISLOWER(c) ('a' <= ((unsigned) (c)) && ((unsigned) (c)) <= 'z')
#define ISALPHA(c) (ISUPPER(c) || ISLOWER(c))
#define ISDIGIT(c) ('0' <= ((unsigned) (c)) && ((unsigned) (c)) <= '9')
#define ISXDIGIT(c) (strchr("abcdefABCDEF", c)!=NULL)
#define ISALNUM(c) (ISALPHA(c) || ISDIGIT(c))
#define ISPRINT(c) (' ' <= ((unsigned) (c)) && ((unsigned) (c)) <= 127)

#define _ARGCV_WORD_SED_EXPR 0x10000
#define _ARGCV_WORD_MASK     0xf0000

#define ALLOC_INIT 128
#define ALLOC_INCR 128

static void
_wsplt_alloc_die(struct wordsplit *wsp)
{
	wsp->ws_error(_("memory exhausted"));
	abort();
}

static void
_wsplt_error(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fputc('\n', stderr);
}

static int
wordsplit_init(struct wordsplit *wsp, const char *input, size_t len,
	       int flags)
{
	wsp->ws_flags = flags;

	if (!(wsp->ws_flags & WRDSF_ALLOC_DIE))
		wsp->ws_alloc_die = _wsplt_alloc_die;
	if (!(wsp->ws_flags & WRDSF_ERROR))
		wsp->ws_error = _wsplt_error;

	if ((wsp->ws_flags & (WRDSF_NOVAR|WRDSF_NOCMD))
	    != (WRDSF_NOVAR|WRDSF_NOCMD)) {
		errno = EINVAL;
		wsp->ws_errno = WRDSE_NOSUPP;
		if (wsp->ws_flags & WRDSF_SHOWERR)
			wordsplit_perror(wsp);
		return wsp->ws_errno;
	}

	wsp->ws_input = input;
	wsp->ws_len = len;

	if (!(wsp->ws_flags & WRDSF_DOOFFS))
		wsp->ws_offs = 0;

	if (!(wsp->ws_flags & WRDSF_DELIM))
		wsp->ws_delim = " ";

	if (!(wsp->ws_flags & WRDSF_COMMENT))
		wsp->ws_comment = NULL;

	if (wsp->ws_flags & WRDSF_REUSE) {
		wsp->ws_wordn = wsp->ws_wordc + 1;
		if (!(wsp->ws_flags & WRDSF_APPEND))
			wsp->ws_wordc = 0;
	} else {
		wsp->ws_wordv = NULL;
		wsp->ws_wordc = 0;
		wsp->ws_wordn = 0;
	}
	if (wsp->ws_flags & WRDSF_DOOFFS)
		wsp->ws_wordn += wsp->ws_offs;

	wsp->ws_endp = 0;
	return 0;
}

static int
alloc_space(struct wordsplit *wsp)
{
	size_t offs = (wsp->ws_flags & WRDSF_DOOFFS) ? wsp->ws_offs : 0;
	char **ptr;
	size_t newalloc;

	if (wsp->ws_wordv == NULL) {
		newalloc = offs + ALLOC_INIT;
		ptr = calloc(newalloc, sizeof (ptr[0]));
	} else if (wsp->ws_wordn < offs + wsp->ws_wordc + 1) {
		newalloc = offs + wsp->ws_wordc + ALLOC_INCR;
		ptr = realloc(wsp->ws_wordv, newalloc * sizeof (ptr[0]));
	}
	else
		return 0;

	if (ptr) {
		wsp->ws_wordn = newalloc;
		wsp->ws_wordv = ptr;
	} else {
		errno = ENOMEM;
		wsp->ws_errno = WRDSE_NOSPACE;
		if (wsp->ws_flags & WRDSF_ENOMEMABRT)
			wsp->ws_alloc_die(wsp);
		else if (wsp->ws_flags & WRDSF_SHOWERR)
			wordsplit_perror(wsp);
		return wsp->ws_errno;
	}
	return 0;
}

static int
skip_sed_expr(const char *command, size_t i, size_t len)
{
	int state;

	do {
		int delim;

		if (command[i] == ';')
			i++;
		if (!(command[i] == 's' && i + 3 < len
		      && ISPUNCT(command[i+1])))
			break;

		delim = command[++i];
		state = 1;
		for (i++; i < len; i++)	{
			if (state == 3) {
				if (command[i] == delim
				    || !ISALNUM(command[i]))
					break;
			} else if (command[i] == '\\')
				i++;
			else if (command[i] == delim)
				state++;
		}
	}
	while (state == 3 && i < len && command[i] == ';');
	return i;
}

static size_t
skip_delim(struct wordsplit *wsp)
{
	size_t start = wsp->ws_endp;
	if (wsp->ws_flags & WRDSF_SQUEEZE_DELIMS) {
		do
			start++;
		while (start < wsp->ws_len
		       && ISDELIM(wsp->ws_input[start], wsp->ws_delim));
		start--;
	}

	if (!(wsp->ws_flags & WRDSF_RETURN_DELIMS))
		start++;

	return start;
}

#define _WRDS_EOF  0
#define _WRDS_WORD 1
#define _WRDS_CONT 2
#define _WRDS_ERR  3

static int
scan_word(struct wordsplit *wsp, size_t *pstart, size_t *pend)
{
	size_t start = *pstart;
	size_t len = wsp->ws_len;
	const char *command = wsp->ws_input;
	const char *delim = wsp->ws_delim;
	const char *comment = wsp->ws_comment;

	size_t i = start;

	if (i >= len) {
		wsp->ws_errno = WRDSE_EOF;
		return _WRDS_EOF;
	}
	
	if (wsp->ws_flags & WRDSF_WS) {
		/* Skip initial whitespace */
		while (ISWS(command[i]))
			if (++i == len) {
				wsp->ws_errno = WRDSE_EOF;
				return _WRDS_EOF;
			}
	}

	start = i;

	wsp->ws_flags &= ~_ARGCV_WORD_MASK;

	if (wsp->ws_flags & WRDSF_SED_EXPR
	    && command[i] == 's' && i + 3 < len && ISPUNCT(command[i+1])) {
		wsp->ws_flags |= _ARGCV_WORD_SED_EXPR;
		i = skip_sed_expr(command, i, len);
	} else if (!ISDELIM(command[i], delim)) {
		while (i < len) {
			if (comment && strchr(comment, command[i]) != NULL) {
				size_t j;
				for (j = i + 1; j < len
					     && command[j] != '\n'; j++)
					;
				*pstart = start;
				*pend = i;
				wsp->ws_endp = j;
				return i > start ? _WRDS_WORD : _WRDS_CONT;
			}

			if (wsp->ws_flags & WRDSF_QUOTE) {
				if (command[i] == '\\')	{
					if (++i == len)
						break;
					i++;
					continue;
				}

				if (command[i] == '\'' || command[i] == '"') {
					size_t j;
					for (j = i + 1; j < len
						     && command[j] != command[i]; j++)
						if (command[j] == '\\')
							j++;
					if (j < len
					    && command[j] == command[i])
						i = j + 1;
					else {
						*pstart = start;
						*pend = i;
						wsp->ws_endp = i;
						wsp->ws_errno = WRDSE_QUOTE;
						if (wsp->ws_flags & WRDSF_SHOWERR)
							wordsplit_perror(wsp);
						return _WRDS_ERR;
					}
				}
			}

			if (((wsp->ws_flags & WRDSF_WS) && ISWS(command[i]))
			    || ISDELIM(command[i], delim))
				break;
			else
				i++;
		}
	} else if (wsp->ws_flags & WRDSF_RETURN_DELIMS)
		i++;

	*pstart = start;
	*pend = i;
	wsp->ws_endp = i;

	return _WRDS_WORD;
}

static char quote_transtab[] = "\\\\a\ab\bf\fn\nr\rt\tv\v";

int
wordsplit_unquote_char(int c)
{
	char *p;

	for (p = quote_transtab; *p; p += 2) {
		if (*p == c)
			return p[1];
	}
	return c;
}

int
wordsplit_quote_char(int c)
{
	char *p;

	for (p = quote_transtab + sizeof(quote_transtab) - 2;
	     p > quote_transtab; p -= 2) {
		if (*p == c)
			return p[-1];
	}
	return -1;
}

#define to_num(c) \
  (ISDIGIT(c) ? c - '0' : (ISXDIGIT(c) ? toupper(c) - 'A' + 10 : 255 ))

static int
xtonum (int *pval, const char *src, int base, int cnt)
{
	int i, val;

	for (i = 0, val = 0; i < cnt; i++, src++) {
		int n = *(unsigned char*)src;
		if (n > 127 || (n = to_num(n)) >= base)
			break;
		val = val * base + n;
	}
	*pval = val;
	return i;
}

size_t
wordsplit_quoted_length(const char *str, int quote_hex, int *quote)
{
	size_t len = 0;

	*quote = 0;
	for (; *str; str++) {
		if (*str == ' ') {
			len++;
			*quote = 1;
		} else if (*str == '"') {
			len += 2;
			*quote = 1;
		} else if (*str != '\t' && *str != '\\' && ISPRINT(*str))
			len++;
		else if (quote_hex)
			len += 3;
		else {
			if (wordsplit_quote_char(*str) != -1)
				len += 2;
			else
				len += 4;
		}
	}
	return len;
}

void
wordsplit_unquote_copy(char *dst, const char *src, size_t n)
{
	int i = 0;
	int c;
	int expect_delim = 0;

	while (i < n) {
		switch (src[i]) {
		case '\'':
		case '"':
			if (!expect_delim) {
				const char *p;

				for (p = src+i+1; *p && *p != src[i]; p++)
					if (*p == '\\')
						p++;
				if (*p)
					expect_delim = src[i++];
				else
					*dst++ = src[i++];
			} else if (expect_delim == src[i])
				++i;
			else
				*dst++ = src[i++];
			break;

		case '\\':
			++i;
			if (src[i] == 'x' || src[i] == 'X') {
				if (n - i < 2) {
					*dst++ = '\\';
					*dst++ = src[i++];
				} else {
					int off = xtonum(&c, src + i + 1,
							 16, 2);
					if (off == 0) {
						*dst++ = '\\';
						*dst++ = src[i++];
					}
					else {
						*dst++ = c;
						i += off + 1;
					}
				}
			} else if ((unsigned char)src[i] < 128
				   && ISDIGIT(src[i])) {
				if (n - i < 1) {
					*dst++ = '\\';
					*dst++ = src[i++];
				} else {
					int off = xtonum(&c, src+i, 8, 3);
					if (off == 0) {
						*dst++ = '\\';
						*dst++ = src[i++];
					} else {
						*dst++ = c;
						i += off;
					}
				}
			} else
				*dst++ = wordsplit_unquote_char(src[i++]);
			break;

		default:
			*dst++ = src[i++];
		}
	}
	*dst = 0;
}

void
wordsplit_quote_copy(char *dst, const char *src, int quote_hex)
{
	for (; *src; src++) {
		if (*src == '"') {
			*dst++ = '\\';
			*dst++ = *src;
		} else if (*src != '\t' && *src != '\\' && ISPRINT(*src))
			*dst++ = *src;
		else {
			char tmp[4];

			if (quote_hex) {
				snprintf(tmp, sizeof tmp, "%%%02X",
					 *(unsigned char*)src);
				memcpy(dst, tmp, 3);
				dst += 3;
			} else {
				int c = wordsplit_quote_char(*src);
				*dst++ = '\\';
				if (c != -1)
					*dst++ = c;
				else {
					snprintf(tmp, sizeof tmp, "%03o",
						 *(unsigned char*)src);
					memcpy(dst, tmp, 3);
					dst += 3;
				}
			}
		}
	}
}

int
wordsplit_len(const char *command, size_t len, struct wordsplit *wsp,
	      int flags)
{
	int rc;
	size_t start = 0, end = 0;

	rc = wordsplit_init(wsp, command, len, flags);
	if (rc)
		return rc;

	for (; (rc = scan_word(wsp, &start, &end)) != _WRDS_EOF;
	     start = skip_delim(wsp)) {
		int unquote = 1;
		size_t n;
		char *p;

		if (rc == _WRDS_CONT)
			continue;
		if (rc == _WRDS_ERR)
			break;

		if (alloc_space(wsp))
			return wsp->ws_errno;

		n = end - start;

		if (wsp->ws_flags & WRDSF_QUOTE &&
		    !(wsp->ws_flags & _ARGCV_WORD_SED_EXPR)) {
			if (start < end
			    && (command[start] == '"'
				|| command[start] == '\'')
			    && command[end-1] == command[start]) {
				unquote = command[start] == '"';
				start++;
				n -= 2;
			}
		} else
			unquote = 0;

		p = malloc(n + 1);
		if (!p) {
			errno = ENOMEM;
			wsp->ws_errno = WRDSE_NOSPACE;
			if (wsp->ws_flags & WRDSF_ENOMEMABRT)
				wsp->ws_alloc_die(wsp);
			if (wsp->ws_flags & WRDSF_SHOWERR)
				wordsplit_perror(wsp);
			if (!(wsp->ws_flags & WRDSF_REUSE))
				wordsplit_free(wsp);
			return wsp->ws_errno;
		}

		if (unquote)
			wordsplit_unquote_copy(p, &command[start], n);
		else {
			memcpy (p, &command[start], n);
			p[n] = 0;
		}
		wsp->ws_wordv[wsp->ws_offs + wsp->ws_wordc] = p;
		wsp->ws_wordc++;
	}
	if (alloc_space (wsp))
		return wsp->ws_errno;
	wsp->ws_wordv[wsp->ws_offs + wsp->ws_wordc] = NULL;
	/* FIXME: if (rc) free(ws) */
	return rc;
}

int
wordsplit(const char *command, struct wordsplit *ws, int flags)
{
	return wordsplit_len(command, strlen (command), ws, flags);
}

void
wordsplit_free(struct wordsplit *ws)
{
	free(ws->ws_wordv);
	ws->ws_wordv = NULL;
}

void
wordsplit_perror(struct wordsplit *wsp)
{
	switch (wsp->ws_errno) {
	case WRDSE_EOF:
		wsp->ws_error(_("no error"));
		break;
		
	case WRDSE_QUOTE:
		wsp->ws_error(_("missing closing %c (start near #%lu)"),
			      wsp->ws_input[wsp->ws_endp], wsp->ws_endp);
		break;

	case WRDSE_NOSPACE:
		wsp->ws_error(_("memory exhausted"));
		break;
		
	case WRDSE_NOSUPP:
		wsp->ws_error(_("variable expansion and "
				"command substitution "
				"are not yet supported"));

	default:
		wsp->ws_error(_("unknown error"));
	}
}

const char *_worsplit_errstr[] = {
	N_("no error"),
	N_("missing closing quote"),
	N_("memory exhausted"),
	N_("variable expansion and command substitution "
	   "are not yet supported")
};
int _worsplit_nerrs = sizeof(_worsplit_errstr)/sizeof(_worsplit_errstr[0]);

const char *
wordsplit_strerror(struct wordsplit *ws)
{
	if (ws->ws_errno < _worsplit_nerrs)
		return _worsplit_errstr[ws->ws_errno];
	return N_("unknown error");
}

